/* 
	HCIEmu - HCI Emulation for Linux Bluetooth protocol stack (BlueZ)
	Copyright (C) 2000-2001 Maxim Krasnyansky <maxk@qualcomm.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 2 as
	published by the Free Software Foundation;

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
	IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY CLAIM,
	OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
	RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
	USE OR PERFORMANCE OF THIS SOFTWARE.

	ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS, COPYRIGHTS,
	TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS SOFTWARE IS DISCLAIMED.
*/
/*
 * $Id: hciemu.c,v 1.3 2002/06/24 03:56:50 maxk Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>
#include <termios.h>
#include <fcntl.h>
#include <ctype.h>

#include <netinet/in.h>
#include <netdb.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <asm/types.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_vhci.h>
#include <bluetooth/l2cap.h>

#include <glib.h>

#include "hciemu.h"

static struct vhci_device vdev;
static struct vhci_conn *vconn[VHCI_MAX_CONN];

static GMainLoop *event_loop;

volatile sig_atomic_t __io_canceled;

gboolean io_acl_data(GIOChannel *chan, GIOCondition cond, gpointer data);
gboolean io_conn_ind(GIOChannel *chan, GIOCondition cond, gpointer data);
gboolean io_hci_data(GIOChannel *chan, GIOCondition cond, gpointer data);

struct vhci_conn * conn_get_by_bdaddr(bdaddr_t *ba)
{
	register int i;

	for (i=0; i < VHCI_MAX_CONN; i++)
		if (!bacmp(&vconn[i]->dest, ba))
			return vconn[i];

	return NULL;
}

void command_status(__u16 ogf, __u16 ocf, __u8 status)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_cmd_status *cs;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_CMD_STATUS;
	he->plen = EVT_CMD_STATUS_SIZE; 

	cs = (void *) ptr; ptr += EVT_CMD_STATUS_SIZE;

	cs->status = status;
	cs->ncmd   = 1;
	cs->opcode = htobs(cmd_opcode_pack(ogf,ocf));
	
	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

void command_complete(__u16 ogf, __u16 ocf, int plen, void *data)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_cmd_complete *cc;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_CMD_COMPLETE;
	he->plen = EVT_CMD_COMPLETE_SIZE + plen; 

	cc = (void *) ptr; ptr += EVT_CMD_COMPLETE_SIZE;

	cc->ncmd = 1;
	cc->opcode = htobs(cmd_opcode_pack(ogf,ocf));
	
	if (plen) {
		memcpy(ptr, data, plen);
		ptr += plen;
	}

	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

void connect_request(struct vhci_conn *conn)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_conn_request *cr;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_CONN_REQUEST;
	he->plen = EVT_CONN_REQUEST_SIZE; 

	cr = (void *) ptr; ptr += EVT_CONN_REQUEST_SIZE;

	bacpy(&cr->bdaddr, &conn->dest);
	memset(&cr->dev_class, 0, sizeof(cr->dev_class));
	cr->link_type = ACL_LINK;
	
	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

void connect_complete(struct vhci_conn *conn)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_conn_complete *cc;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_CONN_COMPLETE;
	he->plen = EVT_CONN_COMPLETE_SIZE; 

	cc = (void *) ptr; ptr += EVT_CONN_COMPLETE_SIZE;

	bacpy(&cc->bdaddr, &conn->dest);
	cc->status = 0;
	cc->handle = htobs(conn->handle);
	cc->link_type = ACL_LINK;
	cc->encr_mode = 0x0;
	
	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

void disconn_complete(struct vhci_conn *conn)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_disconn_complete *dc;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_DISCONN_COMPLETE;
	he->plen = EVT_DISCONN_COMPLETE_SIZE;

	dc = (void *) ptr; ptr += EVT_DISCONN_COMPLETE_SIZE;

	dc->status = 0;
	dc->handle = htobs(conn->handle);
	dc->reason = 0;
	
	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

void num_completed_pkts(struct vhci_conn *conn)
{
	__u8 buf[HCI_MAX_FRAME_SIZE], *ptr = buf;
	evt_num_comp_pkts *np;
	hci_event_hdr *he;

	/* Packet type */
	*ptr++ = HCI_EVENT_PKT;
	
	/* Event header */
	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;

	he->evt  = EVT_NUM_COMP_PKTS;
	he->plen = EVT_NUM_COMP_PKTS_SIZE;

	np = (void *) ptr; ptr += EVT_NUM_COMP_PKTS_SIZE;
	np->num_hndl = 1;

	*((__u16 *)ptr) = htobs(conn->handle);  ptr += 2;
	*((__u16 *)ptr) = htobs(conn->acl_cnt); ptr += 2;

	if (write(vdev.fd, buf, ptr-buf) < 0)
		syslog(LOG_ERR, "Can't send event. %s(%d).", 
					strerror(errno), errno);
}

int scan_enable(__u8 *data)
{
	struct sockaddr_in sa;
	GIOChannel *sk_io;
	bdaddr_t ba;
	int sk, opt;

	if (!(*data & SCAN_PAGE)) {
		if (vdev.scan) {
			g_io_channel_close(vdev.scan);
			vdev.scan = NULL;
		}
		return 0;
	}

	if (vdev.scan)
		return 0;	

	if ((sk = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		syslog(LOG_ERR, "Can't create socket %s(%d)", 
						strerror(errno), errno);
		return 1;
	}

	opt=1;
	setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

	baswap(&ba, &vdev.bdaddr);
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = *(__u32 *)&ba;
	sa.sin_port = *(__u16 *)&ba.b[4];
	if (bind(sk, (struct sockaddr *)&sa, sizeof(sa))) {
		syslog(LOG_ERR, "Can't bind socket %s(%d)", 
						strerror(errno), errno);
		goto failed;
	}
		
	if (listen(sk, 10)) {
		syslog(LOG_ERR, "Can't listen on socket %s(%d)", 
						strerror(errno), errno);
		goto failed;
	}

	sk_io = g_io_channel_unix_new(sk);
	g_io_add_watch(sk_io, G_IO_IN | G_IO_NVAL, io_conn_ind, NULL);
	vdev.scan = sk_io;
	return 0;

failed:
	close(sk);
	return 1;
}

void accept_connection(__u8 *data)
{
	accept_conn_req_cp *cp = (void *)data;
	struct vhci_conn *conn;

	if (!(conn = conn_get_by_bdaddr(&cp->bdaddr)))
		return;

	connect_complete(conn);

	g_io_add_watch(conn->chan, G_IO_IN | G_IO_NVAL | G_IO_HUP, 
			io_acl_data, (gpointer) conn);
}

void close_connection(struct vhci_conn *conn)
{
	syslog(LOG_INFO, "Closing connection %s handle %d", 
					batostr(&conn->dest), conn->handle);

	g_io_channel_close(conn->chan);

	vconn[conn->handle-1] = NULL;
	disconn_complete(conn);
	free(conn);
}

void disconnect(__u8 *data)
{
	disconnect_cp *cp = (void *)data;
	struct vhci_conn *conn;
	__u16 handle;

	handle = btohs(cp->handle);

	if (handle-1 > VHCI_MAX_CONN)
		return;

	if (!(conn = vconn[handle-1]))
		return;

	close_connection(conn);
}

void create_connection(__u8 *data)
{
	create_conn_cp *cp = (void *)data;
	struct vhci_link_info info;
	struct sockaddr_in sa;
	struct vhci_conn *conn;
	int h, sk, opt;
	bdaddr_t ba;

	for (h=0; h < VHCI_MAX_CONN; h++)
		if (!vconn[h])
			goto do_connect;

	syslog(LOG_ERR, "Too many connections");
	return;

do_connect:	
	if ((sk = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		syslog(LOG_ERR, "Can't create socket %s(%d)", 
						strerror(errno), errno);
		return;
	}

	opt=1;
	setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

	baswap(&ba, &vdev.bdaddr);
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY; //*(__u32 *)&ba;
	sa.sin_port = 0; //*(__u16 *)&ba.b[4];
	if (bind(sk, (struct sockaddr *)&sa, sizeof(sa))) {
		syslog(LOG_ERR, "Can't bind socket %s(%d)",
						strerror(errno), errno);
		close(sk);
		return;
	}
	
	baswap(&ba, &cp->bdaddr);
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = *(__u32 *)&ba;
	sa.sin_port = *(__u16 *)&ba.b[4];
	if (connect(sk, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
		syslog(LOG_ERR, "Can't connect %s(%d)", strerror(errno), errno);
		close(sk);
		return;
	}

	/* Send info */
	memset(&info, 0, sizeof(info));
	bacpy(&info.bdaddr, &vdev.bdaddr);
	info.link_type = ACL_LINK;
	info.role = 1;
	write_n(sk, (void *)&info, sizeof(info));

	if (!(conn = malloc(sizeof(*conn)))) {
		syslog(LOG_ERR, "Can't alloc new connection. %s(%d)",
					strerror(errno), errno);
		close(sk);
		return;
	}

	memcpy((__u8 *)&ba, (__u8 *)&sa.sin_addr, 4);
	memcpy((__u8 *)&ba.b[4], (__u8 *)&sa.sin_port, 2);
	baswap(&conn->dest, &ba);

	vconn[h] = conn;
	conn->handle = h+1;
	conn->acl_cnt = 0;
	conn->chan = g_io_channel_unix_new(sk);

	connect_complete(conn);
	g_io_add_watch(conn->chan, G_IO_IN | G_IO_NVAL | G_IO_HUP, 
				io_acl_data, (gpointer) conn);
	return;
}

void inline hci_host_control(__u16 ocf, int plen, __u8 *data)
{
	__u8 status;

	const __u16 ogf = OGF_HOST_CTL;

	switch (ocf) {
	case OCF_RESET:
		status = 0;
		command_complete(ogf, ocf, 1, &status);
		break;

	case OCF_SET_EVENT_FLT:
		status = 0;
		command_complete(ogf, ocf, 1, &status);	
		break;

	case OCF_WRITE_PG_TIMEOUT:
	case OCF_WRITE_CA_TIMEOUT:
		status = 0;
		command_complete(ogf, ocf, 1, &status);	
		break;

	case OCF_WRITE_SCAN_ENABLE:
		status = scan_enable(data);
		command_complete(ogf, ocf, 1, &status);
		break;

	case OCF_WRITE_AUTH_ENABLE:
		status = 0;
		command_complete(ogf, ocf, 1, &status);	
		break;

	default:
		status = 0x01;
		command_complete(ogf, ocf, 1, &status);
		break;
	}
}

void inline hci_link_control(__u16 ocf, int plen, __u8 *data)
{
	__u8 status;

	const __u16 ogf = OGF_LINK_CTL;

	switch (ocf) {
	case OCF_CREATE_CONN:
		command_status(ogf, ocf, 0);
		create_connection(data);
		break;

	case OCF_ACCEPT_CONN_REQ:
		command_status(ogf, ocf, 0);	
		accept_connection(data);
		break;

	case OCF_DISCONNECT:
		command_status(ogf, ocf, 0);
		disconnect(data);
		break;

	default:
		status = 0x01;
		command_complete(ogf, ocf, 1, &status);
		break;
	}
}

void inline hci_link_policy(__u16 ocf, int plen, __u8 *data)
{
	__u8 status;

	const __u16 ogf = OGF_INFO_PARAM;

	switch (ocf) {
	default:
		status = 0x01;
		command_complete(ogf, ocf, 1, &status);
		break;
	}
}

void inline hci_info_param(__u16 ocf, int plen, __u8 *data)
{
	read_local_features_rp lf;
	read_buffer_size_rp bs;
	read_bd_addr_rp ba;
	__u8 status;

	const __u16 ogf = OGF_INFO_PARAM;

	switch (ocf) {
	case OCF_READ_LOCAL_FEATURES:
		lf.status = 0;
		memcpy(lf.features, vdev.features, 8);
		command_complete(ogf, ocf, sizeof(lf), &lf);
		break;

	case OCF_READ_BUFFER_SIZE:
		bs.status = 0;
		bs.acl_mtu = htobs(VHCI_ACL_MTU);
		bs.sco_mtu = 0;
		bs.acl_max_pkt = htobs(VHCI_ACL_MAX_PKT);
		bs.sco_max_pkt = htobs(0);
		command_complete(ogf, ocf, sizeof(bs), &bs);
		break;

	case OCF_READ_BD_ADDR:
		ba.status = 0;
		bacpy(&ba.bdaddr, &vdev.bdaddr);
		command_complete(ogf, ocf, sizeof(ba), &ba);
		break;

	default:
		status = 0x01;
		command_complete(ogf, ocf, 1, &status);
		break;
	}
}

void hci_command(__u8 *data)
{
	hci_command_hdr *ch;
	__u8 *ptr = data;
	__u16 ogf, ocf;

	ch = (hci_command_hdr *)ptr;
	ptr += HCI_COMMAND_HDR_SIZE;

	ch->opcode = btohs(ch->opcode);
	ogf = cmd_opcode_ogf(ch->opcode);
	ocf = cmd_opcode_ocf(ch->opcode);

	switch (ogf) {
	case OGF_HOST_CTL:
		hci_host_control(ocf, ch->plen, ptr);
		break;

	case OGF_INFO_PARAM:
		hci_info_param(ocf, ch->plen, ptr);
		break;
	
	case OGF_LINK_CTL:
		hci_link_control(ocf, ch->plen, ptr);
		break;

	case OGF_LINK_POLICY:
		hci_link_policy(ocf, ch->plen, ptr);
		break;
	}
}

void hci_acl_data(__u8 *data)
{
	hci_acl_hdr *ah = (void *) data;
	struct vhci_conn *conn;
	__u16 handle;
	int fd;

	handle = acl_handle(btohs(ah->handle));

	if (handle > VHCI_MAX_CONN || !(conn = vconn[handle-1])) {
		syslog(LOG_ERR, "Bad connection handle %d", handle);
		return;
	}

	fd = g_io_channel_unix_get_fd(conn->chan);
	if (write_n(fd, data, btohs(ah->dlen) + HCI_ACL_HDR_SIZE) < 0) {
		close_connection(conn);
		return;
	}

	if (++conn->acl_cnt == 10) {
		/* Send num of complete packets event */
		num_completed_pkts(conn);
		conn->acl_cnt = 0;
	}
}

gboolean io_acl_data(GIOChannel *chan, GIOCondition cond, gpointer data) 
{
	struct vhci_conn *conn = (struct vhci_conn *)data;
	char buf[HCI_MAX_FRAME_SIZE], *ptr;
	hci_acl_hdr *ah;
	__u16 flags;
	int len, fd;

	if (cond & G_IO_NVAL)
		return FALSE;

	if (cond & G_IO_HUP) {
		close_connection(conn);
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(chan);

	ptr = buf + 1;
	if (read_n(fd, ptr, HCI_ACL_HDR_SIZE) <= 0) {
		close_connection(conn);
		return FALSE;
	}

	ah = (void *) ptr;
	ptr += HCI_ACL_HDR_SIZE;

	len = btohs(ah->dlen);
	if (read_n(fd, ptr, len) <= 0) { 
		close_connection(conn);
		return FALSE;
	}

	buf[0] = HCI_ACLDATA_PKT;

	flags = acl_flags(btohs(ah->handle));
	ah->handle = htobs(acl_handle_pack(conn->handle, flags));
	len += HCI_ACL_HDR_SIZE + 1;

	write(vdev.fd, buf, len);
	return TRUE;
}

gboolean io_conn_ind(GIOChannel *chan, GIOCondition cond, gpointer data) 
{
	struct vhci_link_info info;
	struct vhci_conn *conn;
	struct sockaddr_in sa;
	int sk, nsk, h;

	if (cond & G_IO_NVAL)
		return FALSE;

	sk = g_io_channel_unix_get_fd(chan);	

	h = sizeof(sa);
	if ((nsk = accept(sk, (struct sockaddr *)&sa, &h)) < 0)
		return TRUE;

	if (read_n(nsk, (void *) &info, sizeof(info)) < 0) {
		syslog(LOG_ERR, "Can't read link info");
		return TRUE;
	}

	if (!(conn = malloc(sizeof(*conn)))) {
		syslog(LOG_ERR, "Can't alloc new connection");
		close(nsk);
		return TRUE;
	}

	bacpy(&conn->dest, &info.bdaddr);

	for (h=0; h < VHCI_MAX_CONN; h++)
		if (!vconn[h])
			goto accepted;

	syslog(LOG_ERR, "Too many connections");
	free(conn);
	close(nsk);
	return TRUE;

accepted:
	vconn[h] = conn;
	conn->handle = h+1;
	conn->acl_cnt = 0;
	conn->chan = g_io_channel_unix_new(nsk);
	connect_request(conn);
	
	return TRUE;
}

gboolean io_hci_data(GIOChannel *chan, GIOCondition cond, gpointer data)
{
	char buf[HCI_MAX_FRAME_SIZE], *ptr;
	int  len, type;
	GIOError err;

	ptr = buf;

	if ((err = g_io_channel_read(chan, buf, sizeof(buf), &len))) {
		if (err == G_IO_ERROR_AGAIN)
			return TRUE;

		syslog(LOG_ERR, "Read failed. %s(%d)", strerror(errno), errno);
		g_main_quit(event_loop);
		return FALSE;
	}

	type = *ptr++;

	switch (type) {
	case HCI_COMMAND_PKT:
		hci_command(ptr);
		break;

	case HCI_ACLDATA_PKT:
		hci_acl_data(ptr);
		break;

	default:
		syslog(LOG_ERR, "Unknown packet type 0x%2.2x.", type);
		break;
	}	
	return TRUE;
}

int getbdaddrbyname(char *str, bdaddr_t *ba)
{
	int i, n, len;

	len = strlen(str);

	/* Check address format */

	for(i=0,n=0; i<len; i++)
		if (str[i] == ':') n++;

	if (n==5) {
		/* BD address */
		baswap(ba, strtoba(str));
		return 0;
	}

	if (n==1) {
		/* IP address + port */
		struct hostent * hent;
		bdaddr_t b;
		char *ptr;

		ptr = strchr(str, ':');
		*ptr++ = 0;

		if (!(hent = gethostbyname(str))) {
			printf("Can't resolve %s\n", str);
			return -2;
		}
		memcpy(&b, hent->h_addr, 4);
		*(__u16 *)(&b.b[4]) = htons(atoi(ptr));
		baswap(ba, &b);
		return 0;
	}

	printf("Invalid address format\n");
	return -1;
}

void sig_term(int sig)
{
	io_cancel();
	g_main_quit(event_loop);
}

void usage(void)
{
	printf("hciemu - HCI emulator ver %s\n", VERSION);
	printf("Usage: \n");
	printf("\thciemu [-n] local_address\n");
}

int main(int argc, char *argv[], char *env[])
{
	extern int optind,opterr,optopt;
	extern char *optarg;
	int daemon, dofork, opt, dev, fd;
	struct sigaction sa;
	GIOChannel *dev_io;
	char *address = NULL;

	/* Configure default settings */
	daemon = 1; dofork = 1;

	while ((opt=getopt(argc,argv,"n")) != EOF) {
		switch(opt) {
			case 'n':
				daemon = 0;
				break;

			default:
				usage();
				exit(1);
		}
	}

	if (argc - optind < 1) {
		usage();
		exit(1);
	}

	address = strdup(argv[optind]);

	if (getbdaddrbyname(address, &vdev.bdaddr) < 0)
		exit(1);

	if (daemon) {
		if (dofork && fork())
			exit(0);

		/* Direct stdin,stdout,stderr to '/dev/null' */
		fd = open("/dev/null", O_RDWR);
		dup2(fd, 0); dup2(fd, 1); dup2(fd, 2);
		close(fd);

		setsid();

		chdir("/");
	}

	/* Start logging to syslog and stderr */
	openlog("hciemu", LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_DAEMON);
	syslog(LOG_INFO, "HCI Emulation daemon ver %s started", VERSION);

	memset(&sa,0,sizeof(sa));
	sa.sa_handler = sig_term;
	sa.sa_flags = SA_NOCLDSTOP;  
	sigaction(SIGTERM, &sa, NULL);

	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	io_init();
	
	/* Device settings */
	vdev.features[0] = 0xff;
	vdev.features[1] = 0xff;
	vdev.features[2] = 0xff;
	vdev.features[3] = 0xff;

	/* Create event loop */
	event_loop = g_main_new(FALSE);

	/* Open and create virtual HCI device */
	if ((dev = open(VHCI_DEV, O_RDWR)) < 0) {
		syslog(LOG_ERR, "Can't open device. %s(%d)", strerror(errno), errno);
		exit(1);
	}
	vdev.fd = dev;
	
	dev_io = g_io_channel_unix_new(dev);
	g_io_add_watch(dev_io, G_IO_IN, io_hci_data, NULL);

	setpriority(PRIO_PROCESS, 0, -19);

	/* Start event processor */
	g_main_run(event_loop);

	syslog(LOG_INFO, "Exit.");
	return 0;
}
